package satstatus

import (
	"log"
	"sync"
	"time"

	"github.com/it-novum/nsta/oitcdb"
)

// Satellite is the common interface of all different satellite types
type Satellite interface {
	SatelliteID() int
}
type Status struct {
	SatelliteID int
	Status      oitcdb.SatelliteStatusID
	LastSeen    time.Time
	LastExport  time.Time
	Err         error
}

type StatusTracker struct {
	db     *oitcdb.OITC
	mapper map[int]*oitcdb.SatelliteStatus
	done   chan struct{}
	wg     sync.WaitGroup
	reload chan map[int]*oitcdb.SatelliteStatus
	Update chan *Status // Channel used by the satellite instances to pass the status information
}

func New(db *oitcdb.OITC) *StatusTracker {
	t := &StatusTracker{
		db:     db,
		done:   make(chan struct{}),
		reload: make(chan map[int]*oitcdb.SatelliteStatus),
		Update: make(chan *Status, 4096),
	}

	t.wg.Add(1)
	go t.run()
	return t
}

// Gets called by and runs in the main thread
func (t *StatusTracker) Reload(satellites []Satellite, reloadBegin time.Time) {
	mapper := map[int]*oitcdb.SatelliteStatus{}
	for _, sat := range satellites {
		i := sat.SatelliteID()
		mapper[i] = t.db.FetchSatelliteStatus(i)
		if reloadBegin.After(mapper[i].LastErrorTime) {
			// Only reset error messages that happend BEFORE the reload happend - not during the reload process
			mapper[i].LastError = ""
			mapper[i].LastErrorTime = time.Time{}
			mapper[i].Save()
		}
	}

	// Pass the new satellite status data to the StatusTracker
	t.reload <- mapper
}

func (t *StatusTracker) run() {
	defer t.wg.Done()
	t.mapper = <-t.reload
	for {
		select {
		case mapper := <-t.reload:
			//Save new satellite status from Reload func
			t.mapper = mapper
		case <-t.done:
			// we have to wait until all data has been flushed
			close(t.Update)
		case status, c := <-t.Update:
			if !c {
				// Channel got closed - exit StatusTracker
				log.Println("shutdown status tracker")
				return
			}
			m := t.mapper[status.SatelliteID]
			if m != nil {
				if !status.LastExport.IsZero() {
					m.LastExport = status.LastExport
				}
				if !status.LastSeen.IsZero() {
					m.LastSeen = status.LastSeen
				}
				if status.Err != nil {
					m.LastError = status.Err.Error()
					m.LastErrorTime = time.Now()
				}
				if status.Status != 0 {
					m.Status = status.Status
				}
				m.Save()
			}
		}
	}
}

func (t *StatusTracker) Close() {
	t.done <- struct{}{}
	t.wg.Wait()
}
